/****************************************************************************
*** Demo program for the remote agent PC and the RAPHAL TSR
*** Copyright (C) 1998 Demmel Hard & Software

*** For MS-C 7.0

HISTORY:
WHEN			WHO	WHAT
----------------------------------------------------------------------------
1998-07-01	Jerry	Creation
1998-11-03	Jerry	New function for entering a line of text
1999-08-17	Jerry	Added functions for LCD contrast, backlight and hotkeys

****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>		/* for register definitions and int86 functions */
#include <bios.h>
#include <conio.h>
#include <ctype.h>
#include <time.h>

/* RAPHAL TSR header file for controlling the TSR */
#include "raphal.h"

/* Booleans */
#define FALSE	0
#define TRUE	(!FALSE)

/* Macro for constructing far pointers */
#ifndef MK_FP
#define MK_FP(s,o) \
	((void far*)(((unsigned long)((unsigned)(s))<<16) | (unsigned)(o)))
#endif

/* Number of array elements */
#define NR_ELEMS(array)	(sizeof(array)/(sizeof(*(array))))

/* Some timer definitions for producing sound */
#define TIMER_CONTROL	0x43
#define TIMER_2_DATA		0x42
#define TIMER_SOUND		182		/* Produce sound */

/* Special bits for enabling/disabling sound */
#define SOUND_ONOFF_PORT	0x61
#define SOUND_ONOFF_BITS	3

/* Convert frequency to timer counts for sound */
#define FREQ(f)	((unsigned)(1193180UL/(f)))

/* Frequency of the error beep */
#define BEEP_FREQ		880
/* Duration of beep in microseconds */
#define BEEP_DURATION	100000UL

/* BIOS variables */
#define BIOS_SEG	0x40	/* Segment number of BIOS variables */
#define BIOS_NR_SCREEN_COLS	0x4A	/* Number of columns of current screen */
#define BIOS_NR_SCREEN_BYTES	0x4C	/* Number of bytes for current screen */

/* Interrupt numbers */
#define VIDEO_INT	0x10	/* Video interrupt hooked by RAPHAL TSR */
#define EXT_INT	0x15	/* "Extended" services */
#define KEYB_INT	0x16	/* Keyboard services */

/* Services of video interrupt */
#define VIDEO_SET_CURSORSIZE	0x01
#define VIDEO_SET_CURSORPOS	0x02
#define VIDEO_SCROLL_UP			0x06
#define VIDEO_WRITE_CHAR		0x0A
#define VIDEO_GET_MODE			0x0F

/* Services of extended services interrupt */
#define EXT_WAIT					0x86

/* Services of keyboard interrupt */
#define KEYB_STUFF_KEY			0x05
#define KEYB_EXT_GET				0x10
#define KEYB_EXT_STATUS			0x11

/* Key code definitions (as returned by INT16H/10H) */
/* We do not care about the ascii code, only normal/extended ("grey") keys.
	This makes us independent from NumLock state and language-specific
	keyboard drivers. */

#define KEY_F1		0x3B00
#define KEY_F2		0x3C00
#define KEY_F3		0x3D00
#define KEY_F4		0x3E00
#define KEY_F5		0x3F00
#define KEY_F6		0x4000
#define KEY_F7		0x4100
#define KEY_F8		0x4200

#define KEY_BS		0x0E00	/* Backspace key ('C' on the Remote Agent PC) */

#define KEY_ESC	0x0100	/* Escape key (not on Remote Agent PC keyboard) */

#define KEY_DOT	0x5300
#define KEY_0		0x5200
#define KEY_1		0x4F00
#define KEY_2		0x5000
#define KEY_3		0x5100
#define KEY_4		0x4B00
#define KEY_5		0x4C00
#define KEY_6		0x4D00
#define KEY_7		0x4700
#define KEY_8		0x4800
#define KEY_9		0x4900
#define KEY_ENTER	0xE000
#define KEY_PLUS	0x4E00
#define KEY_MINUS	0x4A00

#define KEY_UP		0x48E0
#define KEY_DOWN	0x50E0
#define KEY_LEFT	0x4BE0
#define KEY_RIGHT	0x4DE0

/* Mask for getting scancode from key */
#define MASK_SCAN		0xFF00
/* Mask for ASCII code */
#define MASK_ASCII	0x00FF

/* ASCII part of an extended ("grey") key (e.g. cursor keys) */
#define GREY_KEY	0xE0

/* Number of bits for do_bitpattern */
/* Cannot be modified easily! */
#define NR_BITS	8

/* Max number of digits for entering debounce value */
#define DEBOUNCE_LEN	3

/* Max value of debounce value */
#define MAX_DEBOUNCE_VAL	255

/* Max number of digits for the LCD contrast value */
#define CONTRAST_LEN	2

/* Max value of LCD contrast */
#define MAX_CONTRAST_VAL	50

/* Max number of digits for the backlight brightness value */
#define BACKLIGHT_LEN	3

/* Max value of the LCD backlight brightness */
#define MAX_BACKLIGHT_VAL	200

/* Maximum number of digits for function enter_number() */
#define MAX_NR_DIGITS	5

/* Special number returned by enter_number() if user did not enter anything */
#define EN_NONE	0xFFFF

/* Bitmask for some flags in the CPU flags register */
#define CPU_CARRYF	(1 << 0)
#define CPU_ZEROF		(1 << 6)

/* Delay in microseconds between two scroll steps of the intro text */
/* It cannot be much shorter because the latency of LC display makes it
	unreadable otherwise. */
#define INTRO_DELAY	250000UL	/* one-quarter of a second */

/* Bit patterns for leftmost and rightmost LED */
#define LED_LEFTMOST		((unsigned char)'\x01')
#define LED_RIGHTMOST	((unsigned char)'\x80')

/* Line number for scrolling intro text */
#define INTRO_LINE	'\x01'

/* Different cursor shapes accepted by set_cursorsize */
enum cursor_shape_t
{
	cursor_none,		/* No cursor at all */
	cursor_underline,	/* Steady underline */
	cursor_block		/* Blinking block */
};

/* Enum for specifying input or output ports */
enum in_or_out
{
	input_port,
	output_port
};

/* Pointer to a bitpattern processing function */
typedef void (*bitpatfunc_ptr_t)(unsigned bitpattern);

/* Pointer to a menu function */
typedef int (*menufunc_ptr_t)(unsigned key);

/* Forward declaration of menu node type */
struct menunode_t;

/* Structure of a menu item */
struct menuitem_t
{
	unsigned key;	/* Key associated with the menu item */
	menufunc_ptr_t funcptr;		/* Function called when menu item is selected */
	struct menunode_t *menuptr;	/* Submenu when menu item is selected */
};

/* Structure of a menu node */
struct menunode_t
{
	char *menustr;		/* Menu text */
	struct menuitem_t *menuitems;	/* Array of selectable menu items */
};


/* Forward declarations of functions used in the menu tree */
int do_write_output_port(unsigned key);
int do_read_input_port(unsigned key);
int do_set_output_leds(unsigned key);
int do_set_input_leds(unsigned key);
int do_set_debounce_time(unsigned key);
int do_set_output_sync(unsigned key);
int do_set_input_sync(unsigned key);
int do_letters(unsigned key);
int do_reboot(unsigned key);
int do_contrast(unsigned key);
int do_backlight(unsigned key);
int do_cbhotkeys(unsigned key);

/* Screen dimensions set by main() */
int nr_lcd_lines, nr_lcd_cols;

/* Dynamically allocated temp string for write_circular_string() */
char *spart = NULL;

/* Flags if input and output port/LED synchronization is on or off */
int in_sync, out_sync;

/* Title string and scrolling text */
char title_str[] = "REMOTE AGENT PC";
char intro_str[] =
"Demmel prsentiert Ihnen den Remote Agent PC880."
" Drcken Sie eine beliebige Taste (auer FUNC),"
" um in das Men zu gelangen.     ";

/* String for immediate/deferred mode of function do_bitpattern */
/* They must have the same length! */
char mode_imm[] = "Immedia:";		/* Immediate mode */
char mode_def[] = "Deferrd:";		/* Deferred mode */
#define MODE_LEN	(sizeof mode_imm - 1)

/* Translation between scan codes and numerical values on the keypad */
unsigned keypad2num[] =
{
	KEY_0, KEY_1, KEY_2, KEY_3, KEY_4, KEY_5, KEY_6, KEY_7, KEY_8, KEY_9
};

/**** Definition of menu tree ****/
/* Ports menu */
struct menuitem_t port_items[] =
{
	{ KEY_F1, do_write_output_port, NULL },
	{ KEY_F2, do_read_input_port, NULL },
	{ 0, NULL, NULL }
};
struct menunode_t port_menu =
{
	"F1: Write outputF2: Read input",
	port_items
};

/* LEDs menu */
struct menuitem_t led_items[] =
{
	{ KEY_F1, do_set_output_leds, NULL },
	{ KEY_F2, do_set_input_leds, NULL},
	{ 0, NULL, NULL }
};
struct menunode_t led_menu =
{
	"F1: Output LEDs F2: Input LEDs",
	led_items
};

/* Turn output LED synchronization on/off menu */
struct menuitem_t output_sync_onoff_items[] =
{
	{ KEY_F1, do_set_output_sync, NULL },
	{ KEY_F2, do_set_output_sync, NULL },
	{ 0, NULL, NULL }
};
struct menunode_t output_sync_onoff_menu =
{
	"Output LED synchF1: On  F2: Off ",
	output_sync_onoff_items
};

/* Turn input LED synchronization on/off menu */
struct menuitem_t input_sync_onoff_items[] =
{
	{ KEY_F1, do_set_input_sync, NULL },
	{ KEY_F2, do_set_input_sync, NULL },
	{ 0, NULL, NULL }
};
struct menunode_t input_sync_onoff_menu =
{
	"Input LED synch F1: On  F2: Off ",
	input_sync_onoff_items
};

/* Synchronization menu */
struct menuitem_t sync_items[] =
{
	{ KEY_F1, NULL, &output_sync_onoff_menu },
	{ KEY_F2, NULL, &input_sync_onoff_menu },
	{ 0, NULL, NULL }
};
struct menunode_t sync_menu =
{
	"F1: Output LEDs F2: Input LEDs",
	sync_items
};

/* Allow reboot on/off menu */
struct menuitem_t reboot_onoff_items[] =
{
	{ KEY_F1, do_reboot, NULL },
	{ KEY_F2, do_reboot, NULL },
	{ 0, NULL, NULL }
};
struct menunode_t reboot_onoff_menu =
{
	"Allow kbd rebootF1: On  F2: Off ",
	reboot_onoff_items
};

/* Enable/disable LCD hotkeys menu */
struct menuitem_t cbhotkeys_items[] =
{
	{ KEY_F1, do_cbhotkeys, NULL },
	{ KEY_F2, do_cbhotkeys, NULL },
	{ 0, NULL, NULL }
};
struct menunode_t cbhotkeys_menu =
{
	"LCD Hotkeys     F1: On  F2: Off ",
	cbhotkeys_items
};

/* LCD contrast / backlight brightness menu */
struct menuitem_t lcd_items[] =
{
	{ KEY_F1, do_contrast, NULL },
	{ KEY_F2, do_backlight, NULL },
	{ KEY_F3, NULL, &cbhotkeys_menu },
	{ 0, NULL, NULL }
};
struct menunode_t lcd_menu =
{
	"F1:contr|F2:backF3:LCD Hotkeys  ",
	lcd_items
};

/* "More" menu */
struct menuitem_t more_items[] =
{
	{ KEY_F1, do_set_debounce_time, NULL },
	{ KEY_F2, do_letters, NULL },
	{ KEY_F3, NULL, &reboot_onoff_menu },
	{ KEY_F4, NULL, &lcd_menu},
	{ 0, NULL, NULL }
};
struct menunode_t more_menu =
{
	"F1:Debnc|F2:TextF3:Boot |F4:LCD ",
	more_items
};

/* Main menu */
struct menuitem_t main_items[] =
{
	{ KEY_F1, NULL, &port_menu },
	{ KEY_F2, NULL, &led_menu },
	{ KEY_F3, NULL, &sync_menu },
	{ KEY_F4, NULL, &more_menu },
	{ 0, NULL, NULL }
};
struct menunode_t main_menu =
{
	"F1:Ports|F2:LEDsF3:Synch|F4:More",
	main_items
};


/**************** FUNCTIONS *******************/

/* Start sound with "count" period */
void sound_on(unsigned count)
{
	_disable();

	/* Tell timer 2 to produce sound */
	outp(TIMER_CONTROL,TIMER_SOUND);

	/* Set frequency */
	outp(TIMER_2_DATA,(unsigned char)(count & 0xFF));
	outp(TIMER_2_DATA,(unsigned char)(count >> 8));

	_enable();

	/* Turn on sound */
	outp(SOUND_ONOFF_PORT,
		(unsigned char)(inp(SOUND_ONOFF_PORT) | SOUND_ONOFF_BITS));
}


/* Stop sound */
void sound_off(void)
{
	/* Turn off sound */
	outp(SOUND_ONOFF_PORT,
		(unsigned char)(inp(SOUND_ONOFF_PORT) & ~SOUND_ONOFF_BITS));
}


/* Check if RAPHAL TSR is installed */
int is_tsr_installed(void)
{
	union REGS r;

	/* Call multiplexer interrupt with appropriate ID */
	r.x.ax = MUX_ID;
	int86(MUX_INT,&r,&r);

	/* Return TRUE if TSR is installed, FALSE otherwise */
	return r.h.al != 0;
}


/* Switch to LCD or VGA video mode and get screen dimensions */
void lcd_mode(int lcd,int *nr_cols,int *nr_lines)
{
	/* Video mode before switching to LCD */
	/* Value for checking if already set or not */
	static unsigned char prev_mode = (unsigned char)'\xFF';

	/* Pointer to video BIOS variables */
	unsigned far *nr_bytes_ptr = MK_FP(BIOS_SEG,BIOS_NR_SCREEN_BYTES);
	unsigned far *nr_cols_ptr = MK_FP(BIOS_SEG,BIOS_NR_SCREEN_COLS);
	union REGS r;

	/* If switch to previous mode is desired but never switched to LCD,
		return immediately */
	if (!lcd && prev_mode == (unsigned char)'\xFF')
		return;

	/* Get current video mode if switching to LCD */
	if (lcd && prev_mode == (unsigned char)'\xFF')
	{
		r.h.ah = VIDEO_GET_MODE;
		int86(VIDEO_INT,&r,&r);
		prev_mode = r.h.al;
	}

	/* Switch to or from LCD video mode */
	r.x.ax = lcd ? LCD_VIDEO_MODE : prev_mode;
	int86(VIDEO_INT,&r,&r);

	/* Compute screen dimensions */
	if (nr_cols != NULL && nr_lines != NULL)
	{
		*nr_cols = *nr_cols_ptr;

		/* One character needs two bytes (character and attribute) */
		*nr_lines = *nr_bytes_ptr / *nr_cols / 2;
	}
}

/* Set cursor position */
void set_cursorpos(unsigned char col,unsigned char line)
{
	union REGS r;

	r.h.ah = VIDEO_SET_CURSORPOS;
	r.h.bh = 0;	/* Display page must always be zero for the LCD */
	r.h.dh = line;	/* New cursor position */
	r.h.dl = col;
	int86(VIDEO_INT,&r,&r);
}

/* Clear LCD and set cursor to home position */
void clr_lcd(void)
{
	union REGS r;

	/* Use scroll window BIOS service */
	r.h.ah = VIDEO_SCROLL_UP;
	/* Attribute in BH is ignored */
	r.h.al = 0;	/* Zero lines means clear window */
	r.x.cx = 0;	/* Upper left corner */
	r.h.dh = (unsigned char)(nr_lcd_lines - 1);	/* Lower line number */
	r.h.dl = (unsigned char)(nr_lcd_cols - 1);	/* Lower column number */
	int86(VIDEO_INT,&r,&r);

	/* Set cursor to position 0,0 */
	set_cursorpos(0,0);
}

/* Set cursor size */
void set_cursorsize(enum cursor_shape_t shape)
{
	union REGS r;

	/* Set cursor size appropriate for the predefined shapes */
	switch (shape)
	{
		case cursor_none:
			r.h.ch = LCDC_NONE_START;
			r.h.cl = LCDC_NONE_END;
			break;
			
		case cursor_underline:
			r.h.ch = LCDC_UNDERLINE_START;
			r.h.cl = LCDC_UNDERLINE_END;
			break;
			
		case cursor_block:
			r.h.ch = LCDC_BLOCK_START;
			r.h.cl = LCDC_BLOCK_END;
			break;
	}

	r.h.ah = VIDEO_SET_CURSORSIZE;
	int86(VIDEO_INT,&r,&r);
}


/* Returns the character if a key was hit, zero otherwise */
/* Note: the key is not removed from the keyboard buffer */
/* MS-C does not return the state of the zero flag in cflag, so a simple
	int86 cannot be used. */
int key_available(void)
{
#ifdef __TURBOC__
	union REGS r;

	r.h.ah = KEYB_EXT_STATUS;
	int86(KEYB_INT,&r,&r);
	if (r.x.flags & CPU_ZEROF)
		return 0;

	return r.x.ax;
#else
	return _bios_keybrd(_NKEYBRD_READY);
#endif
}


/* Get the next key from the keyboard buffer and wait if there is none */
/* If unascii is true, set the ascii code to zero if not equal to 0xE0 */
unsigned get_key(int unascii)
{
	union REGS r;

	r.h.ah = KEYB_EXT_GET;
	int86(KEYB_INT,&r,&r);

	/* If it is not an extended ("grey") key */
	if (unascii && r.h.al != (unsigned char)'\xE0')
		r.h.al = 0;

	return r.x.ax;
}

/* Wait for the specified number of microseconds */
/* This routine is provided in order not to depend on any compiler-specific
	routines. */
int us_delay(unsigned long us)
{
	union REGS r;

	r.h.ah = EXT_WAIT;
	r.x.cx = (unsigned)(us >> 16);		/* High word of wait time */
	r.x.dx = (unsigned)(us & 0xFFFFUL);	/* Low word of wait time */
	int86(EXT_INT,&r,&r);

	/* If carry flag is set, an error occurred */
	return (r.x.cflag & CPU_CARRYF) == 0;
}


/* A simple error beep */
void beep(void)
{
	sound_on(FREQ(BEEP_FREQ));
	us_delay(BEEP_DURATION);
	sound_off();
}


/* Write a character at current cursor position (without scrolling, cursor
	increment, etc.) cnt times */
void write_char(char c,unsigned cnt)
{
	union REGS r;

	r.h.ah = VIDEO_WRITE_CHAR;
	r.h.al = c;
	r.h.bh = 0;
	r.x.cx = cnt;
	int86(VIDEO_INT,&r,&r);
}


/* Display a string at given position avoiding scrolling if possible */
/*  Note: Does not wrap on the last line! */
void write_string(char *s,unsigned char col,unsigned char line)
{
	set_cursorpos(col,line);

	while (*s)
	{
		/* Output current character */
		write_char(*s,1);

		/* If we are not at end of string, forward cursor */
		if (*++s)
		{
			/* If we reached the end of the current line */
			if (++col >= (unsigned char)nr_lcd_cols)
			{
				/* If we are on the last line */
				if (line == (unsigned char)(nr_lcd_lines - 1))
				{
					/* Stay on the last column of the last line */
					--col;
				}
				else
				{
					/* Go to beginning of next line */
					col = 0;
					++line;
				}
			}
			set_cursorpos(col,line);
		}
	}
}


/* Write a scrolling string (scrolls once per call) */
void write_circular_string(char *s,size_t len,unsigned char line)
{
	static size_t i = 0;	/* Pointer to first character in string to be output
									this time */
	size_t len1st;		/* Length of first part of string */

	/* Check if memory already allocated for temp string */
	if (spart == NULL)
	{
		/* Allocate enough memory for one display line plus '\0' */
		if ((spart = malloc(nr_lcd_cols + 1)) == NULL)
		{
			/* Cannot allocate memory; just panic */
			lcd_mode(FALSE,NULL,NULL);
			fprintf(stderr,"Malloc error in file %s line %d; aborting\n",
				__FILE__,__LINE__);
			exit(1);
		}
	}

	/* Go to first position of given line */
	set_cursorpos(0,line);

	/* Output first part of string until string or display end */
	strncpy(spart,s + i,nr_lcd_cols);
	spart[nr_lcd_cols] = '\0';
	len1st = strlen(spart);
	write_string(spart,0,line);

	/* If string wrapped around to the beginning, output rest */
	if (i + nr_lcd_cols > len)
	{
		strncpy(spart,s,nr_lcd_cols - (len - i));
		spart[nr_lcd_cols - (len - i)] = '\0';
		write_string(spart,(unsigned char)len1st,line);
	}

	/* If we fell off the end of the string, go to the beginning */
	if (++i == len)
		i = 0;
}


/* Enable or disable LED/port synchronization */
void led_port_sync(enum in_or_out port,int sync_em)
{
	union REGS r;

	r.h.al = (unsigned char)(port == input_port ?
		tsr_cmd_dig_sync_in_leds : tsr_cmd_dig_sync_out_leds);
	r.h.ah = TSR_CTRL_ID;
	r.x.bx = sync_em;
	int86(TSR_CTRL_INT,&r,&r);
}


/* Set the specified LEDs to the given value */
void set_leds(enum in_or_out port,unsigned char val)
{
	union REGS r;

	r.h.al = (unsigned char)(port == input_port ?
		tsr_cmd_led_inputs : tsr_cmd_led_outputs);
	r.h.ah = TSR_CTRL_ID;
	r.h.bl = val;
	int86(TSR_CTRL_INT,&r,&r);
}


/* Display title screen until user pressed a key */
int do_title(void)
{
	/* Current state of the leds */
	unsigned char input_leds = LED_LEFTMOST;	/* Leftmost input LED */
	unsigned char output_leds = LED_RIGHTMOST;	/* Rightmost output LED */

	/* Disable synchronization of the LEDs with their associated ports so that
		we can play freely with them */
	led_port_sync(input_port,FALSE);
	led_port_sync(output_port,FALSE);

	/* Disable cursor */
	set_cursorsize(cursor_none);

	/* Output title on first line of display */
	set_cursorpos(0,0);
	printf(title_str);

	/* Display scrolling text */
	while (key_available() == 0)
	{
		/* Scroll the intro text */
		write_circular_string(intro_str,sizeof intro_str - 1,INTRO_LINE);

		/* Let input LED go to the right and output LED to the left */
		set_leds(input_port,input_leds);
		if ((input_leds <<= 1) == 0)
			input_leds = LED_LEFTMOST;
		set_leds(output_port,output_leds);
		if ((output_leds >>= 1) == 0)
			output_leds = LED_RIGHTMOST;

		/* Wait for some time */
		us_delay(INTRO_DELAY);
	}

	/* Enable LED synchronization again */
	/* This will also update the LEDs to match their ports */
	led_port_sync(input_port,TRUE);
	led_port_sync(output_port,TRUE);
	in_sync = out_sync = TRUE;

	/* Fetch key which caused termination */
	return get_key(TRUE) == KEY_ESC;	/* quitflag value */
}


/* Let the user input a bit pattern and call the specified function whenever
	it changes */
int do_bitpattern(bitpatfunc_ptr_t funcptr,unsigned val,unsigned char line)
{
	unsigned key;
	unsigned char i;
	int immediate = TRUE;
	unsigned char col = MODE_LEN;
	unsigned bitmask = 1U;

	/* Output bitpattern */
	if (funcptr != NULL)
		(*funcptr)(val);

	write_string(immediate ? mode_imm : mode_def,0,line);
	for (i = 1; i != 0; i <<= 1)
	{
		set_cursorpos(col++,line);
		write_char((char)(val & i ? '1' : '0'),1);
	}
	col = MODE_LEN;
	set_cursorpos(col,line);

	/* Turn on cursor */
	set_cursorsize(cursor_block);

	while (TRUE)
	{
		/* Get a valid key from the user */
		key = get_key(TRUE);

		/* Megaswitch */
		switch (key)
		{
			case KEY_ESC:
			case KEY_BS:
				/* Turn off cursor again */
				set_cursorsize(cursor_none);
				/* Return appropriate quitflag value */
				return key == KEY_ESC;

			case KEY_0:
				/* If current position is not already zero */
				if (val & bitmask)
				{
					/* Update display */
					write_char('0',1);
					/* Reset bit in value */
					val &= ~bitmask;
					/* Call caller-supplied function */
					if (immediate && funcptr != NULL)
						(*funcptr)(val);
				}
				break;

			case KEY_1:
				/* If current position is not already one */
				if ((val & bitmask) == 0)
				{
					/* Update display */
					write_char('1',1);
					/* Set bit in value */
					val |= bitmask;
					/* Call caller-supplied function */
					if (immediate && funcptr != NULL)
						(*funcptr)(val);
				}
				break;

			case KEY_ENTER:
				/* Call caller-supplied function if not in immediate mode */
				if (!immediate && funcptr != NULL)
					(*funcptr)(val);
				break;

			case KEY_PLUS:
				/* Go to immediate mode */
				if (!immediate)
				{
					immediate = TRUE;
					/* Update mode indicator */
					write_string(mode_imm,0,line);
					/* Restore cursor position */
					set_cursorpos(col,line);
				}
				break;

			case KEY_MINUS:
				/* Go to deferred mode */
				if (immediate)
				{
					immediate = FALSE;
					/* Update mode indicator */
					write_string(mode_def,0,line);
					/* Restore cursor position */
					set_cursorpos(col,line);
				}
				break;

 			case KEY_LEFT:
				/* Cursor left */
				if (--col < MODE_LEN)
					/* Wrap to the right-hand side */
					col = MODE_LEN + NR_BITS - 1;
				/* Move cursor to new position */
				set_cursorpos(col,line);
				/* Update bitmask for setting/resetting bits */
				bitmask = 1 << col - MODE_LEN;
				break;

 			case KEY_RIGHT:
				/* Cursor right */
				if (++col >= MODE_LEN + NR_BITS)
					/* Wrap to the left-hand side */
					col = MODE_LEN;
				/* Move cursor to new position */
				set_cursorpos(col,line);
				/* Update bitmask for setting/resetting bits */
				bitmask = 1 << col - MODE_LEN;
				break;
		}
	}
}


/* Perform actual output operation to output port */
void write_output_port(unsigned val)
{
	union REGS r;

	r.h.al = tsr_cmd_dig_out;
	r.h.ah = TSR_CTRL_ID;
	r.h.bl = (unsigned char)val;
	int86(TSR_CTRL_INT,&r,&r);
}

#ifdef __TURBOC__
#pragma warn -par
#else
#pragma warning(disable:4100)
#endif
/* Output user-specified value to output port */
int do_write_output_port(unsigned key)
{
	/* Output title */
	clr_lcd();
	printf("WRITE OUT PORT");

	/* Let user play with output port */
	return do_bitpattern(write_output_port,0,1);
}
#ifdef __TURBOC__
#pragma warn .par
#else
#pragma warning(default:4100)
#endif

#ifdef __TURBOC__
#pragma warn -par
#else
#pragma warning(disable:4100)
#endif
/* Read input port continuously till key is hit */
int do_read_input_port(unsigned key)
{
	union REGS r;
	unsigned val;
	unsigned char i;
	unsigned prev = 0xFFFFU;

	while (key_available() == 0)
	{
		/* Loop till value changes or user hits a key */
		do
		{
			/* Get current input port state */
			r.h.al = tsr_cmd_dig_in;
			r.h.ah = TSR_CTRL_ID;
			int86(TSR_CTRL_INT,&r,&r);
			val = r.h.bl;
		}
		while (val == prev && key_available() == 0);

		/* Quit loop if user hit a key */
		if (prev == val)
			break;

		prev = val;

		/* Output state of input port in dec, hex and bin */
		clr_lcd();
		printf("INPUT PORT: %3u",val);
		set_cursorpos(0,1);
		printf("%#04x = ",val);
		for (i = 1; i != 0; i <<= 1)
		{
			putchar(val & i ? '1' : '0');
		}
	}

	return get_key(TRUE) == KEY_ESC;	/* quitflag */
}
#ifdef __TURBOC__
#pragma warn .par
#else
#pragma warning(default:4100)
#endif

void set_output_leds(unsigned val)
{
	set_leds(output_port,(unsigned char)val);
}

#ifdef __TURBOC__
#pragma warn -par
#else
#pragma warning(disable:4100)
#endif
int do_set_output_leds(unsigned key)
{
	clr_lcd();

	/* Check if output port/LED are synchronized */
	if (out_sync)
	{
		write_string("Output port/LEDssync is on!<key>",0,0);
		return get_key(TRUE) == KEY_ESC;	/* quitflag */
	}
	else
	{
		/* Output title */
		printf("SET OUT LEDs");

		return do_bitpattern(set_output_leds,0,1);
	}
}
#ifdef __TURBOC__
#pragma warn .par
#else
#pragma warning(default:4100)
#endif

void set_input_leds(unsigned val)
{
	set_leds(input_port,(unsigned char)val);
}

#ifdef __TURBOC__
#pragma warn -par
#else
#pragma warning(disable:4100)
#endif
int do_set_input_leds(unsigned key)
{
	clr_lcd();

	/* Check if input port/LED are synchronized */
	if (in_sync)
	{
		write_string("Input port/LEDs sync is on!<key>",0,0);
		return get_key(TRUE) == KEY_ESC;	/* quitflag */
	}
	else
	{
		/* Output title */
		printf("SET IN LEDs");

		return do_bitpattern(set_input_leds,0,1);
	}
}
#ifdef __TURBOC__
#pragma warn .par
#else
#pragma warning(default:4100)
#endif


/* Function for letting the user enter a numerical value on the second line */
unsigned enter_number(unsigned *value, unsigned max_len, unsigned max_val)
{
	unsigned char col = 0;	/* Current screen column */
	char str[MAX_NR_DIGITS + 1];	/* For entering the value */
	unsigned i;
	unsigned curr_val = 0;
	unsigned key;

	/* Prepare edit string */
	str[0] = '\0';

	/* Let user enter the value on the second line */
	set_cursorpos(0,1);
	set_cursorsize(cursor_block);

	/* Reuse parameter key since we do not need it anyway */
	while ((key = get_key(TRUE)) != KEY_ESC && key != KEY_ENTER)
	{
		switch (key)
		{
			case KEY_BS:	/* Backspace key for deleting last digit */
				if (col > 0)
				{
					/* Delete previous column */
					set_cursorpos(--col,1);
					write_char(' ',1);
					str[col] = '\0';
					curr_val /= 10U;
				}
				break;

			default:
				/* Do nothing if string reached its maximum length */
				if (col < max_len)
				{
					/* Look for numeric key */
					for (i = 0; i < NR_ELEMS(keypad2num); ++i)
					{
						/* If current key matches */
						if (key == keypad2num[i])
						{
							/* Check if value would exceed its maximum value */
							if (curr_val * 10U + i <= max_val)
							{
								/* Value is valid */

								/* Update screen */
								write_char((unsigned char)(i + '0'),1);
								set_cursorpos((unsigned char)(col + 1),1);

								/* Update string */
								str[col++] = (char)(i + '0');
								str[col] = '\0';

								/* Update value */
								curr_val = curr_val * 10U + i;
							}
						}
					}
				}
		}
	}
	/* Turn off cursor again */
	set_cursorsize(cursor_none);

	/* If user entered nothing */
	if (col == 0)
		/* Return special value */
		*value = EN_NONE;
	else
		*value = curr_val;

	return key;
}


int do_set_debounce_time(unsigned key)
{
	unsigned value;

	/* Output title */
	clr_lcd();
	printf("Debounce value?");

	/* Let user enter new value */
	/* Reuse parameter key since we do not need it anyway */
	key = enter_number(&value, DEBOUNCE_LEN, MAX_DEBOUNCE_VAL);

	/* Set new debounce value */
	if (key == KEY_ENTER && value > 0 && value != EN_NONE)
	{
		union REGS r;

		r.h.al = tsr_cmd_dig_debounce;
		r.h.ah = TSR_CTRL_ID;
		r.h.bl = (unsigned char)value;
		int86(TSR_CTRL_INT,&r,&r);
	}

	return key == KEY_ESC;	/* quitflag */
}


int do_set_output_sync(unsigned key)
{
	union REGS r;

	r.h.al = tsr_cmd_dig_sync_out_leds;
	r.h.ah = TSR_CTRL_ID;
	/* F1 is on, F2 is off */
	r.x.bx = key == KEY_F1;
	int86(TSR_CTRL_INT,&r,&r);

	/* Update our own flag */
	out_sync = key == KEY_F1;

	/* Let user know what happened */
	clr_lcd();
	write_string(key == KEY_F1 ? "Output LEDs syncis now on  <key>" :
		"Output LEDs syncis now off <key>",0,0);

	return get_key(TRUE) == KEY_ESC;	/* quitflag */
}

int do_set_input_sync(unsigned key)
{
	union REGS r;

	r.h.al = tsr_cmd_dig_sync_in_leds;
	r.h.ah = TSR_CTRL_ID;
	/* F1 is on, F2 is off */
	r.x.bx = key == KEY_F1;
	int86(TSR_CTRL_INT,&r,&r);

	/* Update our own flag */
	in_sync = key == KEY_F1;

	/* Let user know what happened */
	clr_lcd();
	write_string(key == KEY_F1 ? "Input LEDs sync is now on  <key>" :
		"Input LEDs sync is now off <key>",0,0);

	return get_key(TRUE) == KEY_ESC;	/* quitflag */
}


/* Set level of FUNC mode handling */
void ignore_func(enum func_mode_handling level)
{
	union REGS r;

	r.h.al = tsr_cmd_ignore_func;
	r.h.ah = TSR_CTRL_ID;
	r.x.bx = level;
	int86(TSR_CTRL_INT,&r,&r);
}

/* Let the user enter a line of text */
int do_letters(unsigned key)
{
	unsigned char ascii;		/* ASCII code of current key */
	unsigned char col = 0;	/* Current screen and text string column */

	/* Pointer to entered text */
	char *text_str = malloc(nr_lcd_cols);

	if (!text_str)
	{
		/* Cannot allocate memory; just panic */
		lcd_mode(FALSE,NULL,NULL);
		fprintf(stderr,"Malloc error in file %s line %d; aborting\n",
			__FILE__,__LINE__);
		exit(1);
	}

	/* Initialize text string */
	*text_str = '\0';

	/* Output title */
	clr_lcd();
	printf("Enter some text:");

	/* Go to second line and set a underline cursor so that current letter is
		not obscured */
	set_cursorpos(0,1);
	set_cursorsize(cursor_underline);

	/* Now we need the full FUNC mode handling */
	ignore_func(fmh_full);

	/* Reuse parameter key since we do not need it anyway */
	/* Do not throw away ASCII code because we need it later */
	while (((key = get_key(FALSE)) & MASK_SCAN) != KEY_ESC &&
		(key & MASK_SCAN) != KEY_ENTER)
	{
		switch (key & MASK_SCAN)
		{
			case KEY_BS:	/* Backspace key to delete previous character */
				if (col > 0)
				{
					/* Delete previous column */
					set_cursorpos(--col,1);
					write_char(' ',1);
					text_str[col] = '\0';
				}
				break;

			default:		/* Any other key */
				/* ASCII code is in the low byte of key */
				ascii = (unsigned char)(key & MASK_ASCII);

				/* Beep and ignore if it is not a printable character or we are
					at the end of the string */
				if (iscntrl(ascii) || ascii == GREY_KEY ||
						col == (unsigned char)(nr_lcd_cols - 1))
					beep();
				else
				{
					/* We have a valid key and room for it */
					write_char(ascii, 1);
					set_cursorpos((unsigned char)(col + 1),1);
					text_str[col++] = ascii;
					text_str[col] = '\0';
				}
		}
	}

	/* Back to simple FUNC mode handling */
	ignore_func(fmh_simple);

	/* Turn off cursor again */
	set_cursorsize(cursor_none);

	/* If user hit ENTER, display text again */
	if ((key & MASK_SCAN) == KEY_ENTER)
	{
		clr_lcd();
		printf("You typed:");
		write_string(text_str, 0, 1);

		/* Wait for	any key */
		get_key(FALSE);
	}

	free(text_str);

	return (key & MASK_SCAN) == KEY_ESC;	/* quitflag */
}


/* Allow or deny reboot through the keyboard (Ctrl-Alt-Del which is
	FUNC-F4-C on the Remote Agent PC) */
int do_reboot(unsigned key)
{
	union REGS r;

	r.h.al = tsr_cmd_ignore_reboot;
	r.h.ah = TSR_CTRL_ID;
	/* Allow is F1, deny F2 */
	r.x.bx = key == KEY_F2;
	int86(TSR_CTRL_INT,&r,&r);

	/* Let user know what happened */
	clr_lcd();
	write_string(key == KEY_F1 ? "Keyboard reboot is allowed <key>" :
		"Keyboard reboot is denied <key>",0,0);

	return get_key(TRUE) == KEY_ESC;	/* quitflag */
}


/* Enable or disable the hotkeys for changing the contrast and backlight
	brightness of the LCD */
int do_cbhotkeys(unsigned key)
{
	union REGS r;

	r.h.al = tsr_cmd_cb_hotkeys;
	r.h.ah = TSR_CTRL_ID;
	/* On is F1, off F2 */
	r.x.bx = key == KEY_F1;
	int86(TSR_CTRL_INT,&r,&r);

	/* Let user know what happened */
	clr_lcd();
	write_string(key == KEY_F1 ? "LCD hotkeys are enabled    <key>" :
		"LCD hotkeys are disabled   <key>",0,0);

	return get_key(TRUE) == KEY_ESC;	/* quitflag */
}


/* Let the user enter a value for the LCD contrast */
int do_contrast(unsigned key)
{
	unsigned value;

	/* Output title */
	clr_lcd();
	printf("LCD contrast?");

	/* Let user enter new value */
	/* Reuse parameter key since we do not need it anyway */
	key = enter_number(&value, CONTRAST_LEN, MAX_CONTRAST_VAL);

	/* Set new contrast value */
	if (key == KEY_ENTER && value != EN_NONE)
	{
		union REGS r;

		r.h.al = tsr_cmd_set_contrast;
		r.h.ah = TSR_CTRL_ID;
		r.h.bl = (unsigned char)value;
		int86(TSR_CTRL_INT,&r,&r);
	}

	return key == KEY_ESC;	/* quitflag */
}


/* Let the user enter a value for the LCD backlight brightness */
int do_backlight(unsigned key)
{
	unsigned value;

	/* Output title */
	clr_lcd();
	printf("LCD brightness?");

	/* Let user enter new value */
	/* Reuse parameter key since we do not need it anyway */
	key = enter_number(&value, BACKLIGHT_LEN, MAX_BACKLIGHT_VAL);

	/* Set new contrast value */
	if (key == KEY_ENTER && value != EN_NONE)
	{
		union REGS r;

		r.h.al = tsr_cmd_set_backlight;
		r.h.ah = TSR_CTRL_ID;
		r.h.bl = (unsigned char)value;
		int86(TSR_CTRL_INT,&r,&r);
	}

	return key == KEY_ESC;	/* quitflag */
}


/* Process the menu tree */
int do_menu(struct menunode_t *curr_menu)
{
	/* Flag if redraw of menu title is necessary */
	int redraw = TRUE;

	/* Process current menu until user quits */
	while (TRUE)
	{
		struct menuitem_t *curr_item;
		unsigned key;

		/* Output menu title */
		if (redraw)
		{
			clr_lcd();
			write_string(curr_menu->menustr,0,0);
		}

		/* Get key from user */
		key = get_key(TRUE);

		/* If user pressed escape key, return immediately */
		if (key == KEY_ESC)
			return TRUE;	/* Value for quitflag */
		
		/* If user pressed backspace key, return to previous menu */
		if (key == KEY_BS)
			return FALSE;	/* Value for quitflag */

		/* Call function or menu associated with the key pressed */
		/* A key value of zero marks end of menu item array */
		redraw = FALSE;	/* Just in case the user pressed an illegal key */
		for (curr_item = curr_menu->menuitems;
			curr_item->key != 0;
			++curr_item)
		{
			/* If user pressed key for current menu item */
			if (key == curr_item->key)
			{
				/* If there is a function associated with this item */
				if (curr_item->funcptr != NULL)
				{
					/* Call function with pressed key as parameter */
					if ((*curr_item->funcptr)(key))
						return TRUE;	/* quitflag */
					redraw = TRUE;
				}
				/* If there is a submenu for this item */
				else if (curr_item->menuptr != NULL)
				{
					/* Process submenu */
					if (do_menu(curr_item->menuptr))
						return TRUE;	/* quitflag */
					redraw = TRUE;
				}
				/* Neither a function nor a menu (should never happen) */
			}
		}
	}
}


/* Display date and time continuously */
void signon(void)
{
	static char time_str[61];
	struct tm *ts;
	time_t t;
	time_t old_t = 0;

	printf("%s\n", title_str);

	set_cursorsize(cursor_none);

	while (TRUE)
	{
		time(&t);
		ts = localtime(&t);

		if (old_t != t)
		{
			strftime(time_str, 60, "%m/%d  %H:%M:%S\r", ts);
			printf(time_str);

			old_t = t;
		}

		if (key_available())
		{
			get_key(FALSE);
			break;
		}
	}
}


/*** MAIN ***/
int main(void)
{
	int quitflag = FALSE;

	/* Check if RAPHAL TSR is installed and abort if not */
	if (!is_tsr_installed())
	{
		fprintf(stderr,"The RAPHAL TSR is not installed.\n");
		return 1;
	}

	/* Switch to LCD video mode */
	lcd_mode(TRUE,&nr_lcd_cols,&nr_lcd_lines);

	/*
		Manual refresh is not necessary because switching to LCD video mode
		always refreshes the screen (if in simulation mode).
	*/

	/* Do not handle multikeys because they would interfere with the display */
	ignore_func(fmh_simple);

	signon();					/* show date & time */

	/* Loop till escape is pressed */
	/* Note that there is no escape key on the actual Remote Agent PC
		keyboard */
	do
	{
		/* Clear LCD screen */
		clr_lcd();

		/* Output title and wait for user to hit a key */
		quitflag = do_title();

		/* Process menu tree */
		if (!quitflag)
			quitflag = do_menu(&main_menu);
	}
	while (!quitflag);

	/* Good-bye message */
	clr_lcd();
	printf("Returning to DOSGood bye!");

	/* Enable full FUNC mode handling again (for command line, etc.) */
	ignore_func(fmh_full);

	/* Switch back to previous video mode */
	lcd_mode(FALSE,NULL,NULL);

	/* Deallocate memory for write_circular_string */
	free(spart);

	return 0;
}
