/****************************************************************************
*** Include file for Remote Agent PC Hardware Abstraction Layer
*** Copyright (C) 1998 Demmel Hard & Software

HISTORY:
WHEN			WHO	WHAT
----------------------------------------------------------------------------
1998-07-01	Jerry	Creation V1.0
1998-10-29	Jerry	Removed special LCD characters because RAPHAL now maps
						characters from a specified source character set (default
						is codepage 437)
1999-08-16	Jerry	Added commands for setting LCD backlight brightness and
						contrast
1999-08-17	Jerry	Added command for enabling/disabling backlight and contrast
						hotkeys

****************************************************************************/

#if !defined(__RAPHAL_H)
#define __RAPHAL_H

/**** DEFINES ****/
/* Recommended values for cursor shapes supported by the LCD (INT10H/01H) */
#define LCDC_UNDERLINE_START	11
#define LCDC_UNDERLINE_END		12
#define LCDC_BLOCK_START  		0
#define LCDC_BLOCK_END			13
#define LCDC_NONE_START			31
#define LCDC_NONE_END			31


/* KOMMUNIKATION MIT TSR */
/* Interrupt fuer die Befehle an das TSR; Befehle im AL-Register */
#define TSR_CTRL_INT		0x60

/* TSR-ID im AH-Register fuer Befehle an das TSR, um die Wahrscheinlichkeit
	von Kollisionen mit anderen TSRs zu minimieren */
#define TSR_CTRL_ID		0xF0

/* Multiplexer-Interrupt fuer die Installationsueberpruefung */
#define MUX_INT	0x2F

/* Multiplexer-Kommando in AX fuer die Installationsueberpruefung */
#define MUX_ID		0xC400

/* Video mode for LCD output */
#define LCD_VIDEO_MODE	7	/* Same as monochrome mode */


/**** Typdefinitionen ****/

/* Enum, die die verschiedenen Mglichkeiten der Tastenbehandlung im
	FUNC-Modus enthlt. */
enum func_mode_handling
{
	fmh_full,	/* Alle Mglichkeiten des FUNC-Modus */
	fmh_simple,	/* Multikeys, die nur auf eine andere Taste bersetzt werden
						sollen, werden bersetzt, solche mit mehreren bersetzungen
						werden *unverndert* durchgereicht */
	fmh_filter,	/* Wie voriges, aber Tasten mit mehreren bersetzungen werden
						verworfen */
	fmh_none		/* Der FUNC-Modus kann nicht aktiviert werden */
};

/* TSR-Befehle, die ueber den TSR_CTRL_INT geschickt werden */
/* Der Befehlscode selbst steht im Register AL, im Register AH muss
	zusaetzlich die TSR_CTRL_ID uebergeben werden. Enthaelt AH einen anderen
	Wert, wird die alte Interrupt-Routine aufgerufen. Ist keine solche
	vorhanden, fuehrt das meist zum Absturz des Rechners!
*/
enum tsr_cmds
{
	/* --- Entferne TSR aus dem Speicher ---
	   Parameter: keine
	   Rueckgabe: AX==0: TSR kann nicht entfernt werden;
		sonst: TSR wird bei Programmende entfernt
	*/
	tsr_cmd_uninstall,

	/* --- Digitale Ausgabe ---
	   Parameter: BL: Byte, das ausgegeben werden soll
	   Rueckgabe: keine
	*/
	tsr_cmd_dig_out,

	/* --- Synchronisation zwischen digitalem Ausgabeport und Ausgabe-LEDs
		ein- oder ausschalten ---
	   Parameter: BX==0: Port und LEDs sind unabhaengig;
					  BX!=0: LEDs entsprechen dem aktuellen Zustand des
						      Ausgabeports
		Rueckgabe: keine
	*/
	tsr_cmd_dig_sync_out_leds,

	/* --- Digitale Eingabe ---
		Parameter: keine
		Rueckgabe: BL: Entprellter Zustand des Eingabeports
	*/
	tsr_cmd_dig_in,

	/* --- Synchronisation zwischen digitalem Eingabeport und Eingabe-LEDs
		ein- oder ausschalten ---
	   Parameter: BX==0: Port und LEDs sind unabhaengig;
					  BX!=0: LEDs entsprechen dem entprellten Zustand des
						      Eingabeports
		Rueckgabe: keine
	*/
	tsr_cmd_dig_sync_in_leds,

	/* --- Entprell-Zeit fuer digitalen Eingabeport einstellen ---
		Parameter: BL: Zeit in 55ms-Einheiten, die eine Eingabeleitung stabil
		               sein muss, bevor ihr Zustand akzeptiert wird
							(Standardwert: 2 (110ms))
		Rueckgabe: keine
	*/
	tsr_cmd_dig_debounce,

	/* --- Ausgabe-LEDs setzen ---
		Parameter: BL: LED-Bitmuster: 0: aus, 1: ein; LSB entspricht LED Nr. 1,
		               MSB entspricht LED Nr. 8
		Rueckgabe: keine
	*/
	tsr_cmd_led_outputs,

	/* --- Eingabe-LEDs setzen ---
		Parameter: BL: LED-Bitmuster: 0: aus, 1: ein; LSB entspricht LED Nr. 1,
		               MSB entspricht LED Nr. 8
		Rueckgabe: keine
	*/
	tsr_cmd_led_inputs,

	/* --- Darstellung der simulierten Hardware am Schirm neu aufbauen ---
		Parameter: keine
		Rueckgabe: keine
	*/
	tsr_cmd_sim_refresh,

	/* --- Ignorieren von Ctrl-Alt-Delete (Reboot) aktivieren ---
		Parameter: BX==0: Ctrl-Alt-Delete zulassen (Standard)
					  BX!=0: Ctrl-Alt-Delete ignorieren
		Rueckgabe: keine
		Anmerkung: Ctrl-Alt-Delete wird am Remote Agent PC durch
					  FUNC-F4-C simuliert, die in dieser Reihenfolge gedrckt
					  und gehalten werden mssen.
	*/
	tsr_cmd_ignore_reboot,

	/* --- FUNC-Modus-Behandlung einstellen ---
		Parameter: BX: Eines von enum func_mode_handling
		Rueckgabe: keine
	*/
	tsr_cmd_ignore_func,

	/* --- Hardware-Watchdog ein- oder ausschalten ---
		Parameter: BX==0: Hardware-Watchdog ausschalten
		           BX!=0: Hardware-Watchdog einschalten
		Rueckgabe: keine
	*/
	tsr_cmd_hw_wd_init,

	/* --- Hardware-Watchdog retriggern ---
		Parameter: keine
		Rueckgabe: keine
	*/
	tsr_cmd_hw_wd_retrigger,

	/* --- Helligkeit der LCD-Hintergrundbeleuchtung setzen ---
		Parameter: BL: Helligkeit (0-200; 0...aus, 200...maximale Helligkeit)
		Rueckgabe: keine
	*/
	tsr_cmd_set_backlight,

	/* --- Kontrast des LCDs setzen ---
		Parameter: BL: Kontrast (0-50)
		Rueckgabe: keine
	*/
	tsr_cmd_set_contrast,

	/* --- Hotkeys zum ndern der Hintergrundbeleuchtung und des Kontrasts
	   ein- oder ausschalten ---
		Parameter: BX==0: Hotkeys ausschalten
		           BX!=0: Hotkeys einschalten
		Rueckgabe: keine
	*/
	tsr_cmd_cb_hotkeys
};

#endif

